/*
 * LICENCE[[
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1/CeCILL 2.O 
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the 
 * License. 
 * 
 * The Original Code is utc.fr code. 
 * 
 * The Initial Developer of the Original Code is 
 * david.rivron@utc.fr
 * 
 * Portions created by the Initial Developer are Copyright (C) 2012
 * the Initial Developer. All Rights Reserved. 
 * 
 * Contributor(s): 
 * 
 * 
 * Alternatively, the contents of this file may be used under the terms of 
 * either of the GNU General Public License Version 2 or later (the "GPL"), 
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"), 
 * or the CeCILL Licence Version 2.0 (http://www.cecill.info/licences.en.html), 
 * in which case the provisions of the GPL, the LGPL or the CeCILL are applicable 
 * instead of those above. If you wish to allow use of your version of this file 
 * only under the terms of either the GPL or the LGPL, and not to allow others 
 * to use your version of this file under the terms of the MPL, indicate your 
 * decision by deleting the provisions above and replace them with the notice 
 * and other provisions required by the GPL or the LGPL. If you do not delete 
 * the provisions above, a recipient may use your version of this file under 
 * the terms of any one of the MPL, the GPL, the LGPL or the CeCILL.
 * ]]LICENCE
 */
 
 var util = require('util'),
 	events = require('events'),
	url = require('url');
 	net = require('net'),
 	fs = require('fs'),
	http = require('http'),
	sio = require('socket.io');

/**
 * Serveur de collaboration
 * 
 * Englobe :
 *  - une communication avec le serveur Java (communication "système")
 *  - un serveur HTTP pour l'écoute des requêtes Upgrade (et pouvant servir de proxy au serveur Java)
 *  - les communication WebSocket
 * 
 * Il produit trois évènements à destination des MessagesHandler : init, message, close
 */
var CollabServer = module.exports = function (vOptions) {
	events.EventEmitter.call(this);
	
	if (vOptions.logPath) {
		var vLogStream = fs.createWriteStream(vOptions.logPath);
		Object.defineProperty(process, 'stdout', { value: vLogStream });
		Object.defineProperty(process, 'stderr', { value: vLogStream });
	}
	/*
	 * Initialisation du serveur
	 */
	var vSelf = this;
	var vClients = this.clients = {};
	this.options = vOptions;
	
	if (vOptions.systemClientPort) {
		// Connexion au serveur Java
		var vSystemSocket = net.connect(vOptions.systemClientPort, function() {
			createServer(vSystemSocket);
		});
		
		// Lorsque le serveur est client du serveur Java, sa durée de vie correspond à celle du socket de communication
		vSystemSocket.on('close', function () {
			process.exit(0)
		});
	} else if (vOptions.systemServerPort) {
		// Ecoute de connexion en provenance du serveur Java
		net.createServer.call(this, function (pSystemSocket) {
			createServer(pSystemSocket);
		}).listen(vOptions.systemServerPort);
				
		if (vOptions.systemConnectUrl) {
			var vConnectRequest = http.request(url.parse(vOptions.systemConnectUrl));
			vConnectRequest.on('error', function(pError) { });
			vConnectRequest.end();
		}
	}
		
	function createServer (pSystemSocket) {
		// Initialisation de l'écoute des messages systèmes
		pSystemSocket.on('data', function (pRawMessage) {
			try {
				var vMessage = JSON.parse(pRawMessage);
			} catch (e) {
				console.error('Invalid system message:', pRawMessage.toString('UTF-8'));
				return;
			}
			vSelf.onSystemMessage(vMessage, pSystemSocket);
		});
	
		/*
		 * Initialisation du serveur HTTP
		 */
		var vHttpServer = http.createServer(function(pRequest, pResponse) {
			if (vOptions.proxyPort) {
				vSelf.onProxyRequest (pRequest, pResponse);
			} else {
				pResponse.statusCode = 404;
				pRequest.on('end', function () {
					pResponse.end();
				});
			}
		}).listen(vOptions.port, vOptions.host || '127.0.0.1');
		
		// L'URL pour la connection WebSocket est envoyé au serveur Java lorsque le serveur HTTP est prêt
		vHttpServer.on('listening', function () {
			var vAddress = vHttpServer.address();
			var vWsUrl = 'ws://' + vAddress.address + ':' + vAddress.port + vWsPath;
			pSystemSocket.write(JSON.stringify({ type: 'context', wsUrl: vWsUrl}));
		});
		
		/*
		 * Initialisation du support des WebSockets sur le serveur HTTP
		 */
		var vWsPath = vOptions.contextPath ? vOptions.contextPath : "/";
		var vWsServer = sio.listen(vHttpServer, {
			transports: [ 'websocket', 'xhr-polling' ],
			'resource': vWsPath,
			'close timeout': 300,
			'heartbeats': false,
			'browser client': false,
			'log level': vOptions.logDebug ? 3 : 2
		}).sockets;
		
		// Connexion d'un client
		vWsServer.on('connection', function (pSocket) {
			// Si il n'y a pas de message init avant une seconde, le client est déconnecté
			var noInitTimeout = setTimeout(function() {
				pSocket.disconnect();
			}, 1000);
			
			// Message init : initialisation du client et vérification des permissions initiales
			pSocket.on('init', function(pMessage) {
				vSelf.logMessage(pMessage);
				var vClient = vClients[pMessage.clientId];
				if (vClient && vClient.token == pMessage.token) {
				 	clearTimeout(noInitTimeout);
					delete vClient.token;
				 	vClient.socket = pSocket;
				 	vClient.pendingWspPerms = {};
				 	vClient.wsps = [];
				 	if (pMessage.wsps.length) {
						// Vérification des droits sur les wsp
				 		pMessage.wsps.forEach(function (pWspCd) {
				 			vClient.pendingWspPerms[pWspCd] = [ pMessage ];
					 		pSystemSocket.write(JSON.stringify({ type: 'checkPerm', wspCd: pWspCd, account: vClient.account, clientId: vClient.clientId }));
				 		})
				 	} else {
				 		vSelf.emit('init', vClient, pMessage);
				 	}
				} else {
					console.error('init message discarded:', pMessage);
					pSocket.disconnect();
				}
				vSelf.logClients(vClients);
			});
			
						
			pSocket.on('message', function (pMessage) {
				vSelf.logMessage(pMessage);
				var vClient = vClients[pMessage.clientId];
				if (vClient && vClient.socket == pSocket) vSelf.emit('message', vClient, pMessage);
				else console.error('uri message discarded:', pMessage);
				vSelf.logClients(vClients);
			});

			pSocket.on('disconnect', function() {
				for (var vClientId in vClients) {
					if (vClients[vClientId].socket == pSocket) {
						vSelf.emit('close', vClients[vClientId]);
						delete vClients[vClientId];
						vSelf.logClients(vClients);
						break;
					}
				}
			});	
		});

	}
}

util.inherits(CollabServer, events.EventEmitter);

CollabServer.prototype.onSystemMessage = function (pMessage, pSystemSocket) {
	/**
	 * Handler des messages systèmes (provenant du serveur Java par l'entrée standard)
	 */
	var vSelf = this;
	if (pMessage.type == 'newClient') {
		// L'utilisateur est ajouté à la liste globale
		delete pMessage.type;
		var vClient = this.clients[pMessage.clientId] = pMessage;
		vClient.urisStates = {};
		pSystemSocket.write(JSON.stringify({ type: 'ack' }));
	} else if (pMessage.type == 'hasPerm') {
		var vClient = this.clients[pMessage.clientId];
		if (pMessage.value) {
			vClient.wsps.push(pMessage.wspCd);
			var vPendingWspPerms = vClient.pendingWspPerms[pMessage.wspCd];
			delete vClient.pendingWspPerms[pMessage.wspCd];
			// Traitement des messages en attente de la permission
			vPendingWspPerms.forEach(function (pMess) {
				if (pMess.type == 'init') {
					vSelf.emit('init', vClient, pMess);
				} else if (pMess.type == 'uri') {
					vSelf.emit('message', vClient, pMess);
				}
			});
		}
		
	}
	this.logClients(this.clients);
}
	
/**
 * Handler d'une requête à rediriger vers le serveur Java
 * 
 * La redirection est "brute" : les évènements d'entrée/sortie sont redirigés tel quel.
 * Il n'y a pas de modification des en-têtes autre que Host qui est reconstruit.
 */
CollabServer.prototype.onProxyRequest = function (pRequest, pResponse) {
	delete pRequest.headers['host'];
	var vProxyRequestParams = {
		port: this.options.proxyPort,
		host: 'localhost',
		method: pRequest.method,
		path: pRequest.url,
		headers: pRequest.headers
	}
	var vProxyRequest = http.request(vProxyRequestParams, function (vProxyResponse) {
		vProxyResponse.on('data', function(chunk) {
			pResponse.write(chunk, 'binary');
		});
		vProxyResponse.on('end', function() {
			pResponse.end();
		});
		pResponse.writeHead(vProxyResponse.statusCode, vProxyResponse.headers);
	});
	pRequest.on('data', function(chunk) {
		vProxyRequest.write(chunk, 'binary');
	});
	pRequest.on('end', function() {
		vProxyRequest.end();
	});
},
	
CollabServer.prototype.logMessage = function (pMessage, pDest) {
	if (this.options.logDebug) console.info(pDest ? '-> ' + pDest : '<-', pMessage);
}

CollabServer.prototype.logClients = function () {
	if (this.options.logDebug) console.info('clients:', this.clients);
}



